# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import shutil
import tempfile
import unittest

from swh.core import config


class ConfReaderTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # create a temporary folder
        cls.tmpdir = tempfile.mkdtemp(prefix='test-swh-core.')
        cls.conffile = os.path.join(cls.tmpdir, 'config.ini')
        conf_contents = """[main]
a = 1
b = this is a string
c = true
h = false
ls = list, of, strings
li = 1, 2, 3, 4
"""
        with open(cls.conffile, 'w') as conf:
            conf.write(conf_contents)

        cls.non_existing_conffile = os.path.join(cls.tmpdir,
                                                 'config-nonexisting.ini')

        # Create an unreadable, proper configuration file
        cls.perms_broken_file = os.path.join(cls.tmpdir, 'unreadable.ini')
        with open(cls.perms_broken_file, 'w') as conf:
            conf.write(conf_contents)
        os.chmod(cls.perms_broken_file, 0o000)

        # Create a proper configuration file in an unreadable directory
        cls.perms_broken_dir = os.path.join(cls.tmpdir, 'unreadabledir')
        cls.file_in_broken_dir = os.path.join(cls.perms_broken_dir,
                                              'unreadable.ini')
        os.makedirs(cls.perms_broken_dir)
        with open(cls.file_in_broken_dir, 'w') as conf:
            conf.write(conf_contents)
        os.chmod(cls.perms_broken_dir, 0o000)

        cls.empty_conffile = os.path.join(cls.tmpdir, 'empty.ini')
        open(cls.empty_conffile, 'w').close()

        cls.default_conf = {
            'a': ('int', 2),
            'b': ('string', 'default-string'),
            'c': ('bool', True),
            'd': ('int', 10),
            'e': ('int', None),
            'f': ('bool', None),
            'g': ('string', None),
            'h': ('bool', True),
            'i': ('bool', True),
            'ls': ('list[str]', ['a', 'b', 'c']),
            'li': ('list[int]', [42, 43]),
        }

        cls.other_default_conf = {
            'a': ('int', 3),
        }

        cls.full_default_conf = cls.default_conf.copy()
        cls.full_default_conf['a'] = cls.other_default_conf['a']

        cls.parsed_default_conf = {
            key: value
            for key, (type, value)
            in cls.default_conf.items()
        }

        cls.parsed_conffile = {
            'a': 1,
            'b': 'this is a string',
            'c': True,
            'd': 10,
            'e': None,
            'f': None,
            'g': None,
            'h': False,
            'i': True,
            'ls': ['list', 'of', 'strings'],
            'li': [1, 2, 3, 4],
        }

    @classmethod
    def tearDownClass(cls):
        # Make the broken perms items readable again to be able to remove them
        os.chmod(cls.perms_broken_dir, 0o755)
        os.chmod(cls.perms_broken_file, 0o644)
        shutil.rmtree(cls.tmpdir)

    def test_read(self):
        # when
        res = config.read(self.conffile, self.default_conf)

        # then
        self.assertEqual(res, self.parsed_conffile)

    def test_read_empty_file(self):
        # when
        res = config.read(None, self.default_conf)

        # then
        self.assertEqual(res, self.parsed_default_conf)

    def test_support_non_existing_conffile(self):
        # when
        res = config.read(self.non_existing_conffile, self.default_conf)

        # then
        self.assertEqual(res, self.parsed_default_conf)

    def test_support_empty_conffile(self):
        # when
        res = config.read(self.empty_conffile, self.default_conf)

        # then
        self.assertEqual(res, self.parsed_default_conf)

    def test_raise_on_broken_directory_perms(self):
        with self.assertRaises(PermissionError):
            config.read(self.file_in_broken_dir, self.default_conf)

    def test_raise_on_broken_file_perms(self):
        with self.assertRaises(PermissionError):
            config.read(self.perms_broken_file, self.default_conf)

    def test_merge_default_configs(self):
        # when
        res = config.merge_default_configs(self.default_conf,
                                           self.other_default_conf)

        # then
        self.assertEqual(res, self.full_default_conf)

    def test_priority_read_nonexist_conf(self):
        # when
        res = config.priority_read([self.non_existing_conffile, self.conffile],
                                   self.default_conf)

        # then
        self.assertEqual(res, self.parsed_conffile)

    def test_priority_read_conf_nonexist_empty(self):
        # when
        res = config.priority_read([
            self.conffile,
            self.non_existing_conffile,
            self.empty_conffile,
        ], self.default_conf)

        # then
        self.assertEqual(res, self.parsed_conffile)

    def test_priority_read_empty_conf_nonexist(self):
        # when
        res = config.priority_read([
            self.empty_conffile,
            self.conffile,
            self.non_existing_conffile,
        ], self.default_conf)

        # then
        self.assertEqual(res, self.parsed_default_conf)

    def test_swh_config_paths(self):
        res = config.swh_config_paths('foo/bar.ini')

        self.assertEqual(res, [
            '~/.config/swh/foo/bar.ini',
            '~/.swh/foo/bar.ini',
            '/etc/softwareheritage/foo/bar.ini',
        ])

    def test_prepare_folder(self):
        # given
        conf = {'path1': os.path.join(self.tmpdir, 'path1'),
                'path2': os.path.join(self.tmpdir, 'path2', 'depth1')}

        # the folders does not exists
        self.assertFalse(os.path.exists(conf['path1']),
                         "path1 should not exist.")
        self.assertFalse(os.path.exists(conf['path2']),
                         "path2 should not exist.")

        # when
        config.prepare_folders(conf, 'path1')

        # path1 exists but not path2
        self.assertTrue(os.path.exists(conf['path1']),
                        "path1 should now exist!")
        self.assertFalse(os.path.exists(conf['path2']),
                         "path2 should not exist.")

        # path1 already exists, skips it but creates path2
        config.prepare_folders(conf, 'path1', 'path2')

        self.assertTrue(os.path.exists(conf['path1']),
                        "path1 should still exist!")
        self.assertTrue(os.path.exists(conf['path2']),
                        "path2 should now exist.")



# Copyright (C) 2015-2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
import functools
import inspect
import json
import logging
import pickle
import requests

from flask import Flask, Request, Response
from .serializers import (decode_response,
                          encode_data_client as encode_data,
                          msgpack_dumps, msgpack_loads, SWHJSONDecoder)


class RemoteException(Exception):
    pass


def remote_api_endpoint(path):
    def dec(f):
        f._endpoint_path = path
        return f
    return dec


class MetaSWHRemoteAPI(type):
    """Metaclass for SWHRemoteAPI, which adds a method for each endpoint
    of the database it is designed to access.

    See for example :class:`swh.indexer.storage.api.client.RemoteStorage`"""
    def __new__(cls, name, bases, attributes):
        # For each method wrapped with @remote_api_endpoint in an API backend
        # (eg. :class:`swh.indexer.storage.IndexerStorage`), add a new
        # method in RemoteStorage, with the same documentation.
        #
        # Note that, despite the usage of decorator magic (eg. functools.wrap),
        # this never actually calls an IndexerStorage method.
        backend_class = attributes.get('backend_class', None)
        for base in bases:
            if backend_class is not None:
                break
            backend_class = getattr(base, 'backend_class', None)
        if backend_class:
            for (meth_name, meth) in backend_class.__dict__.items():
                if hasattr(meth, '_endpoint_path'):
                    cls.__add_endpoint(meth_name, meth, attributes)
        return super().__new__(cls, name, bases, attributes)

    @staticmethod
    def __add_endpoint(meth_name, meth, attributes):
        wrapped_meth = inspect.unwrap(meth)

        @functools.wraps(meth)  # Copy signature and doc
        def meth_(*args, **kwargs):
            # Match arguments and parameters
            post_data = inspect.getcallargs(
                    wrapped_meth, *args, **kwargs)

            # Remove arguments that should not be passed
            self = post_data.pop('self')
            post_data.pop('cur', None)
            post_data.pop('db', None)

            # Send the request.
            return self.post(meth._endpoint_path, post_data)
        attributes[meth_name] = meth_


class SWHRemoteAPI(metaclass=MetaSWHRemoteAPI):
    """Proxy to an internal SWH API

    """

    backend_class = None
    """For each method of `backend_class` decorated with
    :func:`remote_api_endpoint`, a method with the same prototype and
    docstring will be added to this class. Calls to this new method will
    be translated into HTTP requests to a remote server.

    This backend class will never be instantiated, it only serves as
    a template."""

    def __init__(self, api_exception, url, timeout=None):
        super().__init__()
        self.api_exception = api_exception
        base_url = url if url.endswith('/') else url + '/'
        self.url = base_url
        self.session = requests.Session()
        self.timeout = timeout

    def _url(self, endpoint):
        return '%s%s' % (self.url, endpoint)

    def raw_post(self, endpoint, data, **opts):
        if self.timeout and 'timeout' not in opts:
            opts['timeout'] = self.timeout
        try:
            return self.session.post(
                self._url(endpoint),
                data=data,
                **opts
            )
        except requests.exceptions.ConnectionError as e:
            raise self.api_exception(e)

    def raw_get(self, endpoint, params=None, **opts):
        if self.timeout and 'timeout' not in opts:
            opts['timeout'] = self.timeout
        try:
            return self.session.get(
                self._url(endpoint),
                params=params,
                **opts
            )
        except requests.exceptions.ConnectionError as e:
            raise self.api_exception(e)

    def post(self, endpoint, data, params=None):
        data = encode_data(data)
        response = self.raw_post(
            endpoint, data, params=params,
            headers={'content-type': 'application/x-msgpack'})
        return self._decode_response(response)

    def get(self, endpoint, params=None):
        response = self.raw_get(endpoint, params=params)
        return self._decode_response(response)

    def post_stream(self, endpoint, data, params=None):
        if not isinstance(data, collections.Iterable):
            raise ValueError("`data` must be Iterable")
        response = self.raw_post(endpoint, data, params=params)
        return self._decode_response(response)

    def get_stream(self, endpoint, params=None, chunk_size=4096):
        response = self.raw_get(endpoint, params=params, stream=True)
        return response.iter_content(chunk_size)

    def _decode_response(self, response):
        if response.status_code == 404:
            return None
        if response.status_code == 500:
            data = decode_response(response)
            if 'exception_pickled' in data:
                raise pickle.loads(data['exception_pickled'])
            else:
                raise RemoteException(data['exception'])

        # XXX: this breaks language-independence and should be
        # replaced by proper unserialization
        if response.status_code == 400:
            raise pickle.loads(decode_response(response))
        elif response.status_code != 200:
            raise RemoteException(
                "Unexpected status code for API request: %s (%s)" % (
                    response.status_code,
                    response.content,
                )
            )
        return decode_response(response)


class BytesRequest(Request):
    """Request with proper escaping of arbitrary byte sequences."""
    encoding = 'utf-8'
    encoding_errors = 'surrogateescape'


def encode_data_server(data):
    return Response(
        msgpack_dumps(data),
        mimetype='application/x-msgpack',
    )


def decode_request(request):
    content_type = request.mimetype
    data = request.get_data()

    if content_type == 'application/x-msgpack':
        r = msgpack_loads(data)
    elif content_type == 'application/json':
        r = json.loads(data, cls=SWHJSONDecoder)
    else:
        raise ValueError('Wrong content type `%s` for API request'
                         % content_type)

    return r


def error_handler(exception, encoder):
    # XXX: this breaks language-independence and should be
    # replaced by proper serialization of errors
    logging.exception(exception)
    response = encoder(pickle.dumps(exception))
    response.status_code = 400
    return response


class SWHServerAPIApp(Flask):
    """For each endpoint of the given `backend_class`, tells app.route to call
    a function that decodes the request and sends it to the backend object
    provided by the factory.

    :param Any backend_class: The class of the backend, which will be
                              analyzed to look for API endpoints.
    :param Callable[[], backend_class] backend_factory: A function with no
                                                        argument that returns
                                                        an instance of
                                                        `backend_class`."""
    request_class = BytesRequest

    def __init__(self, *args, backend_class=None, backend_factory=None,
                 **kwargs):
        super().__init__(*args, **kwargs)

        if backend_class is not None:
            if backend_factory is None:
                raise TypeError('Missing argument backend_factory')
            for (meth_name, meth) in backend_class.__dict__.items():
                if hasattr(meth, '_endpoint_path'):
                    self.__add_endpoint(meth_name, meth, backend_factory)

    def __add_endpoint(self, meth_name, meth, backend_factory):
        from flask import request

        @self.route('/'+meth._endpoint_path, methods=['POST'])
        @functools.wraps(meth)  # Copy signature and doc
        def _f():
            # Call the actual code
            obj_meth = getattr(backend_factory(), meth_name)
            return encode_data_server(obj_meth(**decode_request(request)))



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import logging
import os
import socket

import psycopg2
from psycopg2.extras import Json
from systemd.journal import JournalHandler as _JournalHandler, send

try:
    from celery import current_task
except ImportError:
    current_task = None


EXTRA_LOGDATA_PREFIX = 'swh_'


def db_level_of_py_level(lvl):
    """convert a log level of the logging module to a log level suitable for the
    logging Postgres DB

    """
    return logging.getLevelName(lvl).lower()


def get_extra_data(record, task_args=True):
    """Get the extra data to insert to the database from the logging record"""
    log_data = record.__dict__

    extra_data = {k[len(EXTRA_LOGDATA_PREFIX):]: v
                  for k, v in log_data.items()
                  if k.startswith(EXTRA_LOGDATA_PREFIX)}

    args = log_data.get('args')
    if args:
        extra_data['logging_args'] = args

    # Retrieve Celery task info
    if current_task and current_task.request:
        extra_data['task'] = {
            'id': current_task.request.id,
            'name': current_task.name,
        }
        if task_args:
            extra_data['task'].update({
                'kwargs': current_task.request.kwargs,
                'args': current_task.request.args,
            })

    return extra_data


def flatten(data, separator='_'):
    """Flatten the data dictionary into a flat structure"""
    def inner_flatten(data, prefix):
        if isinstance(data, dict):
            for key, value in data.items():
                yield from inner_flatten(value, prefix + [key])
        elif isinstance(data, (list, tuple)):
            for key, value in enumerate(data):
                yield from inner_flatten(value, prefix + [str(key)])
        else:
            yield prefix, data

    for path, value in inner_flatten(data, []):
        yield separator.join(path), value


def stringify(value):
    """Convert value to string"""
    if isinstance(value, datetime.datetime):
        return value.isoformat()

    return str(value)


class PostgresHandler(logging.Handler):
    """log handler that store messages in a Postgres DB

    See swh-core/swh/core/sql/log-schema.sql for the DB schema.

    All logging methods can be used as usual. Additionally, arbitrary metadata
    can be passed to logging methods, requesting that they will be stored in
    the DB as a single JSONB value. To do so, pass a dictionary to the 'extra'
    kwarg of any logging method; all keys in that dictionary that start with
    EXTRA_LOGDATA_PREFIX (currently: ``swh_``) will be extracted to form the
    JSONB dictionary. The prefix will be stripped and not included in the DB.

    Note: the logger name will be used to fill the 'module' DB column.

    Sample usage::

        logging.basicConfig(level=logging.INFO)
        h = PostgresHandler('dbname=softwareheritage-log')
        logging.getLogger().addHandler(h)

        logger.info('not so important notice',
                    extra={'swh_type': 'swh_logging_test',
                           'swh_meditation': 'guru'})
        logger.warn('something weird just happened, did you see that?')

    """

    def __init__(self, connstring):
        """
        Create a Postgres log handler.

        Args:
            config: configuration dictionary, with a key "log_db" containing a
               libpq connection string to the log DB
        """
        super().__init__()

        self.connstring = connstring

        self.fqdn = socket.getfqdn()  # cache FQDN value

    def _connect(self):
        return psycopg2.connect(self.connstring)

    def emit(self, record):
        msg = self.format(record)
        extra_data = get_extra_data(record)

        if 'task' in extra_data:
            task_args = {
                'args': extra_data['task']['args'],
                'kwargs': extra_data['task']['kwargs'],
            }

            try:
                json_args = Json(task_args).getquoted()
            except TypeError:
                    task_args = {
                        'args': ['<failed to convert arguments to JSON>'],
                        'kwargs': {},
                    }
            else:
                json_args_length = len(json_args)
                if json_args_length >= 1000:
                    task_args = {
                        'args': ['<arguments too long>'],
                        'kwargs': {},
                    }

            extra_data['task'].update(task_args)

        log_entry = (db_level_of_py_level(record.levelno), msg,
                     Json(extra_data), record.name, self.fqdn,
                     os.getpid())
        db = self._connect()
        with db.cursor() as cur:
            cur.execute('INSERT INTO log '
                        '(level, message, data, src_module, src_host, src_pid)'
                        'VALUES (%s, %s, %s, %s, %s, %s)',
                        log_entry)
            db.commit()
        db.close()


class JournalHandler(_JournalHandler):
    def emit(self, record):
        """Write `record` as a journal event.

        MESSAGE is taken from the message provided by the user, and PRIORITY,
        LOGGER, THREAD_NAME, CODE_{FILE,LINE,FUNC} fields are appended
        automatically. In addition, record.MESSAGE_ID will be used if present.
        """
        try:
            extra_data = flatten(get_extra_data(record, task_args=False))
            extra_data = {
                (EXTRA_LOGDATA_PREFIX + key).upper(): stringify(value)
                for key, value in extra_data
            }
            msg = self.format(record)
            pri = self.mapPriority(record.levelno)
            send(msg,
                 PRIORITY=format(pri),
                 LOGGER=record.name,
                 THREAD_NAME=record.threadName,
                 CODE_FILE=record.pathname,
                 CODE_LINE=record.lineno,
                 CODE_FUNC=record.funcName,
                 **extra_data)
        except Exception:
            self.handleError(record)



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import os
import unittest

import pytest

from swh.core.logger import PostgresHandler
from swh.core.tests.db_testing import SingleDbTestFixture

from swh.core.tests import SQL_DIR


@pytest.mark.db
class PgLogHandler(SingleDbTestFixture, unittest.TestCase):

    TEST_DB_DUMP = os.path.join(SQL_DIR, 'log-schema.sql')

    def setUp(self):
        super().setUp()
        self.modname = 'swh.core.tests.test_logger'
        self.logger = logging.Logger(self.modname, logging.DEBUG)
        self.logger.addHandler(PostgresHandler('dbname=' + self.TEST_DB_NAME))

    def tearDown(self):
        logging.shutdown()
        super().tearDown()

    def test_log(self):
        self.logger.info('notice',
                         extra={'swh_type': 'test entry', 'swh_data': 42})
        self.logger.warning('warning')

        with self.conn.cursor() as cur:
            cur.execute('SELECT level, message, data, src_module FROM log')
            db_log_entries = cur.fetchall()

        self.assertIn(('info', 'notice', {'type': 'test entry', 'data': 42},
                       self.modname),
                      db_log_entries)
        self.assertIn(('warning', 'warning', {}, self.modname), db_log_entries)
